<?php
namespace App\Http;

use App\Entity;
use App\Exception;
use App\Exception\NotLoggedInException;
use App\Exception\PermissionDeniedException;
use App\Service\Sentry;
use App\Session\Flash;
use App\Settings;
use App\View;
use Gettext\Translator;
use Mezzio\Session\SessionInterface;
use Monolog\Logger;
use Psr\Http\Message\ResponseInterface;
use Psr\Http\Message\ServerRequestInterface;
use Psr\Log\LogLevel;
use Slim\App;
use Slim\Exception\HttpException;
use stdClass;
use Whoops\Handler\PrettyPageHandler;
use Whoops\Run;

class ErrorHandler extends \Slim\Handlers\ErrorHandler
{
    protected bool $returnJson = false;

    protected bool $showDetailed = false;

    protected string $loggerLevel = LogLevel::ERROR;

    protected Router $router;

    protected View $view;

    protected Sentry $sentry;

    protected Settings $settings;

    public function __construct(
        App $app,
        Logger $logger,
        Router $router,
        View $view,
        Sentry $sentry,
        Settings $settings
    ) {
        parent::__construct($app->getCallableResolver(), $app->getResponseFactory(), $logger);
        
        $this->settings = $settings;
        $this->router = $router;
        $this->view = $view;
        $this->sentry = $sentry;
    }

    public function __invoke(
        ServerRequestInterface $request,
        \Throwable $exception,
        bool $displayErrorDetails,
        bool $logErrors,
        bool $logErrorDetails
    ): ResponseInterface {
        if ($exception instanceof \App\Exception) {
            $this->loggerLevel = $exception->getLoggerLevel();
        } elseif ($exception instanceof HttpException) {
            $this->loggerLevel = LogLevel::WARNING;
        }

        $this->showDetailed = (!$this->settings->isProduction() && !in_array($this->loggerLevel,
                [LogLevel::DEBUG, LogLevel::INFO, LogLevel::NOTICE], true));
        $this->returnJson = $this->_shouldReturnJson($request);

        return parent::__invoke($request, $exception, $displayErrorDetails, $logErrors, $logErrorDetails);
    }

    /**
     * @return bool
     */
    public function returnJson(): bool
    {
        return $this->returnJson;
    }

    /**
     * @param bool $returnJson
     */
    public function setReturnJson(bool $returnJson): void
    {
        $this->returnJson = $returnJson;
    }

    /**
     * @return bool
     */
    public function showDetailed(): bool
    {
        return $this->showDetailed;
    }

    /**
     * @param bool $showDetailed
     */
    public function setShowDetailed(bool $showDetailed): void
    {
        $this->showDetailed = $showDetailed;
    }

    /**
     * @param ServerRequestInterface $req
     *
     * @return bool
     */
    protected function _shouldReturnJson(ServerRequestInterface $req): bool
    {
        $xhr = $req->getHeaderLine('X-Requested-With') === 'XMLHttpRequest';

        if ($xhr || $this->settings->isCli() || $this->settings->isTesting()) {
            return true;
        }

        if ($req->hasHeader('Accept')) {
            $accept = $req->getHeader('Accept');
            if (in_array('application/json', $accept)) {
                return true;
            }
        }

        return false;
    }

    /**
     * @inheritDoc
     */
    protected function writeToErrorLog(): void
    {
        $context = [
            'file' => $this->exception->getFile(),
            'line' => $this->exception->getLine(),
            'code' => $this->exception->getCode(),
        ];

        if ($this->exception instanceof Exception) {
            $context['context'] = $this->exception->getLoggingContext();
            $context = array_merge($context, $this->exception->getExtraData());
        }

        if ($this->showDetailed) {
            $context['trace'] = array_slice($this->exception->getTrace(), 0, 5);
        }

        $this->logger->log($this->loggerLevel, $this->exception->getMessage(), [
            'file' => $this->exception->getFile(),
            'line' => $this->exception->getLine(),
            'code' => $this->exception->getCode(),
        ]);
    }

    protected function respond(): ResponseInterface
    {
        if (!function_exists('__')) {
            $translator = new Translator();
            $translator->register();
        }

        // Special handling for cURL requests.
        $ua = $this->request->getHeaderLine('User-Agent');

        if (false !== stripos($ua, 'curl')) {
            $response = $this->responseFactory->createResponse($this->statusCode);

            $response->getBody()
                ->write('Error: ' . $this->exception->getMessage() . ' on ' . $this->exception->getFile() . ' L' . $this->exception->getLine());

            return $response;
        }

        $this->view->addData([
            'request' => $this->request,
        ]);

        if ($this->exception instanceof HttpException) {
            /** @var Response $response */
            $response = $this->responseFactory->createResponse($this->exception->getCode());

            if ($this->returnJson) {
                return $response->withJson(new Entity\Api\Error(
                    $this->exception->getCode(),
                    $this->exception->getMessage()
                ));
            }

            return $this->view->renderToResponse(
                $response,
                'system/error_http',
                [
                    'exception' => $this->exception,
                ]
            );
        }

        if ($this->exception instanceof NotLoggedInException) {
            /** @var SessionInterface $session */
            $session = $this->request->getAttribute(ServerRequest::ATTR_SESSION);

            /** @var Flash $flash */
            $flash = $this->request->getAttribute(ServerRequest::ATTR_SESSION_FLASH);

            /** @var Response $response */
            $response = $this->responseFactory->createResponse(403);

            $error_message = __('You must be logged in to access this page.');

            if ($this->returnJson) {
                return $response->withJson(new Entity\Api\Error(403, $error_message));
            }

            // Redirect to login page for not-logged-in users.
            $flash->addMessage(__('You must be logged in to access this page.'), Flash::ERROR);

            // Set referrer for login redirection.
            $session->set('login_referrer', $this->request->getUri()->getPath());

            return $response->withRedirect((string)$this->router->named('account:login'));
        }

        if ($this->exception instanceof PermissionDeniedException) {
            /** @var Flash $flash */
            $flash = $this->request->getAttribute(ServerRequest::ATTR_SESSION_FLASH);

            /** @var Response $response */
            $response = $this->responseFactory->createResponse(403);

            $error_message = __('You do not have permission to access this portion of the site.');

            if ($this->returnJson) {
                return $response->withJson(new Entity\Api\Error(403, $error_message));
            }

            // Bounce back to homepage for permission-denied users.
            $flash->addMessage(__('You do not have permission to access this portion of the site.'),
                Flash::ERROR);

            return $response->withRedirect((string)$this->router->named('home'));
        }

        if (!in_array($this->loggerLevel, [LogLevel::INFO, LogLevel::DEBUG, LogLevel::NOTICE], true)) {
            $this->sentry->handleException($this->exception);
        }

        /** @var Response $response */
        $response = $this->responseFactory->createResponse(500);

        if ($this->returnJson) {
            if ($this->showDetailed) {
                return $response->withJson([
                    'code' => $this->exception->getCode(),
                    'message' => $this->exception->getMessage(),
                    'file' => $this->exception->getFile(),
                    'line' => $this->exception->getLine(),
                    'trace' => $this->exception->getTrace(),
                ]);
            }

            $api_response = new Entity\Api\Error(
                $this->exception->getCode(),
                $this->exception->getMessage(),
                ($this->exception instanceof Exception) ? $this->exception->getFormattedMessage() : $this->exception->getMessage()
            );

            return $response->withJson($api_response);
        }

        if ($this->showDetailed && class_exists('\Whoops\Run')) {
            // Register error-handler.
            $handler = new PrettyPageHandler;
            $handler->setPageTitle('An error occurred!');

            if ($this->exception instanceof Exception) {
                $extra_tables = $this->exception->getExtraData();
                foreach ($extra_tables as $legend => $data) {
                    $handler->addDataTable($legend, $data);
                }
            }

            $run = new Run;
            $run->prependHandler($handler);

            return $response->write($run->handleException($this->exception));
        }

        return $this->view->renderToResponse(
            $response,
            'system/error_general',
            [
                'exception' => $this->exception,
            ]
        );
    }

    /**
     * @param int $code
     * @param string $message
     * @param array $stack_trace
     *
     * @return stdClass
     */
    protected function getErrorApiResponse($code = 500, $message = 'General Error', $stack_trace = []): stdClass
    {
        $api = new stdClass;
        $api->success = false;
        $api->code = (int)$code;
        $api->message = (string)$message;
        $api->stack_trace = (array)$stack_trace;

        return $api;
    }

    protected function withJson(ResponseInterface $response, $data): ResponseInterface
    {
        $json = (string)json_encode($data);
        $response->getBody()->write($json);

        return $response->withHeader('Content-Type', 'application/json;charset=utf-8');
    }
}
