<?php
namespace App\Sync\Task;

use App\Console\Application;
use App\Entity;
use App\Message;
use App\MessageQueue;
use Cake\Chronos\Chronos;
use Doctrine\ORM\EntityManager;
use Psr\Log\LoggerInterface;

class Backup extends AbstractTask
{
    public const BASE_DIR = '/var/azuracast/backups';

    protected MessageQueue $messageQueue;

    protected Application $console;

    public function __construct(
        EntityManager $em,
        Entity\Repository\SettingsRepository $settingsRepo,
        LoggerInterface $logger,
        MessageQueue $message_queue,
        Application $console
    ) {
        parent::__construct($em, $settingsRepo, $logger);

        $this->messageQueue = $message_queue;
        $this->console = $console;
    }

    /**
     * Handle event dispatch.
     *
     * @param Message\AbstractMessage $message
     */
    public function __invoke(Message\AbstractMessage $message)
    {
        if ($message instanceof Message\BackupMessage) {

            $this->settingsRepo->setSetting(Entity\Settings::BACKUP_LAST_RUN, time());

            [$result_code, $result_output] = $this->runBackup(
                $message->path,
                $message->excludeMedia,
                $message->outputPath
            );

            $this->settingsRepo->setSetting(Entity\Settings::BACKUP_LAST_RESULT, $result_code);
            $this->settingsRepo->setSetting(Entity\Settings::BACKUP_LAST_OUTPUT, $result_output);
        }
    }

    /**
     * @param string|null $path
     * @param bool $excludeMedia
     * @param string|null $outputPath
     *
     * @return array [$result_code, $result_output]
     */
    public function runBackup(?string $path = null, bool $excludeMedia = false, ?string $outputPath = null): array
    {
        $input_params = [];
        if (null !== $path) {
            $input_params['path'] = $path;
        }
        if ($excludeMedia) {
            $input_params['--exclude-media'] = true;
        }

        return $this->console->runCommandWithArgs(
            'azuracast:backup',
            $input_params,
            $outputPath ?? 'php://temp'
        );
    }

    /**
     * @inheritdoc
     */
    public function run($force = false): void
    {
        $backup_enabled = (bool)$this->settingsRepo->getSetting(Entity\Settings::BACKUP_ENABLED, 0);
        if (!$backup_enabled) {
            $this->logger->debug('Automated backups disabled; skipping...');
            return;
        }

        $now_utc = Chronos::now('UTC');

        $threshold = $now_utc->subDay()->getTimestamp();
        $last_run = $this->settingsRepo->getSetting(Entity\Settings::BACKUP_LAST_RUN, 0);

        if ($last_run <= $threshold) {
            // Check if the backup time matches (if it's set).
            $backupTimecode = (int)$this->settingsRepo->getSetting(Entity\Settings::BACKUP_TIME);
            
            if (0 !== $backupTimecode) {
                $isWithinTimecode = false;
                $backupDt = Entity\StationSchedule::getDateTime($backupTimecode, $now_utc);

                /** @var Chronos[] $backupTimesToCheck */
                $backupTimesToCheck = [
                    $backupDt->subDay(),
                    $backupDt,
                ];

                foreach ($backupTimesToCheck as $backupStart) {
                    $backupEnd = $backupStart->addMinutes(15);

                    if ($now_utc->between($backupStart, $backupEnd)) {
                        $isWithinTimecode = true;
                        break;
                    }
                }

                if (!$isWithinTimecode) {
                    return;
                }
            }

            // Trigger a new backup.
            $message = new Message\BackupMessage;
            $message->path = 'automatic_backup.zip';
            $message->excludeMedia = (bool)$this->settingsRepo->getSetting(Entity\Settings::BACKUP_EXCLUDE_MEDIA, 0);
            $this->messageQueue->produce($message);
        }
    }
}
